from typing import BinaryIO

import tomli

from pkonfig.storage.base import AbstractStorage, BaseFileStorage, MODE


class Toml(BaseFileStorage, AbstractStorage):
    mode: MODE = "rb"

    def load_file_content(self, handler: BinaryIO) -> None:  # type: ignore
        self.data.update(tomli.load(handler))



import logging
from decimal import Decimal
from enum import Enum
from pathlib import Path
from typing import Any, Callable, Generic, Optional, Sequence, Type, TypeVar

from pkonfig.base import NOT_SET, Field


class Bool(Field):
    def cast(self, value) -> bool:
        return bool(value)


class Int(Field):
    def cast(self, value) -> int:
        return int(value)


class Float(Field):
    def cast(self, value) -> float:
        return float(value)


class DecimalField(Field):
    def cast(self, value) -> Decimal:
        return Decimal(float(value))


class Str(Field):
    def cast(self, value) -> str:
        return str(value)


class Byte(Field):
    def cast(self, value) -> bytes:
        return bytes(value)


class ByteArray(Field):
    def cast(self, value) -> bytearray:
        return bytearray(value)


class PathField(Field):
    value: Path
    missing_ok: bool

    def __init__(self, default=NOT_SET, no_cache=False, missing_ok=False):
        self.missing_ok = missing_ok
        super().__init__(default, no_cache)

    def cast(self, value) -> Path:
        return Path(value)

    def validate(self, value: Path) -> None:
        if not value.exists() and not self.missing_ok:
            raise FileNotFoundError(f"File {value.absolute()} not found")


class File(PathField):
    def validate(self, value):
        if (value.exists() and value.is_file()) or self.missing_ok:
            return
        raise TypeError(f"{value.absolute()} is not a file")


class Folder(PathField):
    def validate(self, value):
        if (value.exists() and value.is_dir()) or self.missing_ok:
            return
        raise TypeError(f"{value.absolute()} is not a directory")


class EnumField(Field):
    def __init__(self, enum_cls: Type[Enum], default=NOT_SET, no_cache=False):
        self.enum_cls = enum_cls
        super().__init__(default, no_cache)

    def cast(self, value: str) -> Enum:
        return self.enum_cls[value]


class LogLevel(Field):
    class Levels(Enum):
        NOTSET = logging.NOTSET
        DEBUG = logging.DEBUG
        INFO = logging.INFO
        WARNING = logging.WARNING
        ERROR = logging.ERROR
        CRITICAL = logging.CRITICAL

    def cast(self, value: str) -> int:
        return self.Levels[value.upper()].value


T = TypeVar("T")


class Choice(Field, Generic[T]):
    def __init__(
        self,
        choices: Sequence[T],
        cast_function: Optional[Callable[[Any], T]] = None,
        default=NOT_SET,
        no_cache=False,
    ):
        self.choices = choices
        self.cast_function = cast_function
        super().__init__(default, no_cache)

    def cast(self, value: T) -> T:
        if self.cast_function is not None:
            value = self.cast_function(value)
        return value

    def validate(self, value):
        if value not in self.choices:
            raise TypeError(f"'{value}' is not in {self.choices}")


class DebugFlag(Field):
    def cast(self, value: str) -> T:
        return value.lower() == "true"



from decimal import Decimal
from pathlib import Path
from typing import Any, Dict, Optional, Type

from pkonfig.base import (
    BaseOuterConfig,
    BaseConfig,
    NOT_SET,
    TypeMapper,
    Field,
)
from pkonfig.fields import (
    Bool,
    Int,
    Float,
    Str,
    Byte,
    ByteArray,
    PathField,
    DecimalField,
)


class DefaultMapper(TypeMapper):
    type_mapping: Dict[Type, Type[Field]] = {
        bool: Bool,
        int: Int,
        float: Float,
        str: Str,
        bytes: Byte,
        bytearray: ByteArray,
        Path: PathField,
        Decimal: DecimalField,
    }

    def descriptor(self, type_: Type, value: Any = NOT_SET) -> Field:
        try:
            cls = self.type_mapping[type_]
            return cls(value)
        except KeyError:
            return value


class Config(BaseOuterConfig):
    _mapper = DefaultMapper()


class EmbeddedConfig(BaseConfig):
    _mapper = DefaultMapper()

    def __init__(self, alias: Optional[str] = None):
        self._alias = alias



from pkonfig.base import (
    Field,
    TypeMapper,
    BaseConfig,
    BaseOuterConfig,
)
from pkonfig.storage import *
from pkonfig.fields import (
    Bool,
    Int,
    Float,
    DecimalField,
    Str,
    Byte,
    ByteArray,
    PathField,
    File,
    Folder,
    EnumField,
    LogLevel,
    Choice,
    DebugFlag,
)
from pkonfig.config import DefaultMapper, Config, EmbeddedConfig



from pkonfig.storage.base import (
    AbstractStorage,
    BaseFileStorage,
    PlainStructureParserMixin,
    Env,
    DotEnv,
    Json,
    Ini,
)

__all__ = [
    "AbstractStorage",
    "BaseFileStorage",
    "PlainStructureParserMixin",
    "Env",
    "DotEnv",
    "Json",
    "Ini",
]

try:
    from pkonfig.storage.toml import Toml
    __all__.append("Toml")
except ImportError:
    pass

try:
    from pkonfig.storage.yaml_ import Yaml
    __all__.append("Yaml")
except ImportError:
    pass



from typing import IO

import yaml

from pkonfig.storage.base import AbstractStorage, BaseFileStorage


class Yaml(BaseFileStorage, AbstractStorage):
    def load_file_content(self, handler: IO) -> None:
        self.data.update(yaml.safe_load(handler))



from abc import ABC, ABCMeta, abstractmethod
from inspect import isclass, isdatadescriptor, ismethod
from typing import (
    Any,
    Dict,
    Generic,
    Iterable,
    Mapping,
    Optional,
    Type,
    TypeVar,
    Union,
    get_type_hints,
)


T = TypeVar("T")
NOT_SET = object()


class Field(ABC, Generic[T]):
    default: Union[T, object]
    value: Union[T, object]
    nullable: bool

    def __init__(
        self,
        default=NOT_SET,
        no_cache=False,
        alias: Optional[str] = None,
        nullable=False,
    ):
        self.no_cache = no_cache
        self.default: Union[T, object] = default
        self.value: Union[T, object] = NOT_SET
        self.alias = alias
        self.nullable = default is None or nullable

    def __set_name__(self, _, name):
        self.name = self.alias if self.alias is not None else name

    def __set__(self, _, value):
        value = self.cast(value)
        self.validate(value)
        self.value = value

    def __get__(self, instance: "BaseConfig", _=None) -> Union[T, object]:
        if self.should_get_from_storage():
            value = self.get_from_storage(instance)
            if value is not None:
                value = self.cast(value)
                self.validate(value)
            self.value = value
        return self.value

    def get_from_storage(self, instance: "BaseConfig") -> Any:
        storage = instance.get_storage()
        value = storage.get(self.name, self.default)
        if value is None and not self.nullable:
            raise ValueError(f"{self.name} is not nullable")
        elif value is NOT_SET:
            raise KeyError(self.name)
        return value

    def should_get_from_storage(self) -> bool:
        return self.value is NOT_SET or self.no_cache

    @abstractmethod
    def cast(self, value: Any) -> T:
        pass

    def validate(self, value: Any) -> None:
        pass


TypeMapping = Dict[Type, Type[Field]]


class TypeMapper(ABC):
    type_mapping: TypeMapping = {}

    def __init__(self, mappings: Optional[TypeMapping] = None):
        if mappings:
            self.type_mapping.update(mappings)

    def replace_fields_with_descriptors(
        self, attributes: Dict[str, Any], type_hints: Dict[str, Type]
    ) -> None:
        for name in filter(lambda x: not x.startswith("_"), type_hints.keys()):
            attribute = attributes.get(name, NOT_SET)
            if self.replace(attribute):
                hint = type_hints[name]
                descriptor = self.descriptor(hint, attribute)
                attributes[name] = descriptor

    @staticmethod
    def replace(attribute):
        return not (isdatadescriptor(attribute) or isclass(attribute))

    @abstractmethod
    def descriptor(self, type_: Type, value: Any = NOT_SET) -> Field:
        pass


class MetaConfig(ABCMeta):
    def __new__(mcs, name, parents, attributes):
        MetaConfig.extend_annotations(attributes)
        mapper = MetaConfig.get_mapper(attributes, parents)
        if mapper:
            mapper.replace_fields_with_descriptors(
                attributes, attributes.get("__annotations__", {})
            )
        cls = super().__new__(mcs, name, parents, attributes)
        return cls

    @staticmethod
    def extend_annotations(attributes: Dict[str, Any]) -> None:
        annotations = attributes.get("__annotations__", {})
        for name, attr in attributes.items():
            if not name.startswith("_") or isclass(attr):
                if name not in annotations:
                    attr_type = type(attr)
                    if issubclass(attr_type, Field):
                        annotation = get_type_hints(attr.cast).get("return", Any)
                    else:
                        annotation = attr_type
                    annotations[name] = annotation
        attributes["__annotations__"] = annotations

    @staticmethod
    def get_mapper(
        attributes: Dict[str, Any], parents: Iterable[Type]
    ) -> Optional[TypeMapper]:
        for name, attribute in attributes.items():
            if isinstance(attribute, TypeMapper):
                return attribute
        for cls in parents:
            mapper = getattr(cls, "_mapper", None)
            if mapper and isinstance(mapper, TypeMapper):
                return mapper
        return None


class BaseConfig(metaclass=MetaConfig):
    _mapper: TypeMapper
    _storage: Mapping
    _alias: Optional[str] = None

    def get_storage(self) -> Mapping:
        return self._storage

    def set_storage(self, storage: Mapping) -> None:
        self._storage = storage

    def is_user_attr(self, name: str) -> bool:
        if name.startswith("_"):
            return False

        if hasattr(self, name):
            attribute = getattr(self, name)
            return not (ismethod(attribute) or isclass(attribute))

        return True

    def user_fields(self):
        return filter(self.is_user_attr, get_type_hints(self).keys())

    def check_all_fields(self):
        for name in self.user_fields():
            attr = getattr(self, name)
            if isinstance(attr, BaseConfig):
                alias = attr._alias if attr._alias else name
                attr.set_storage(self.get_storage().get(alias, {}))
                attr.check_all_fields()


class BaseOuterConfig(BaseConfig):
    def __init__(self, storage: Mapping):
        self.set_storage(storage)
        self.check_all_fields()



import json
import os
from abc import ABC, abstractmethod
from collections import UserDict
from pathlib import Path
from typing import Any, Dict, List, Tuple, IO, Union
import configparser

try:
    from typing import Literal
except ImportError:
    from typing_extensions import Literal   # type:ignore


MODE = Literal["r", "rb"]


class AbstractStorage(UserDict, ABC):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.load()

    @abstractmethod
    def load(self) -> None:
        """Load storage from source to cache"""
        pass


class BaseFileStorage(AbstractStorage, ABC):
    file: Union[Path, str]
    mode: MODE = "r"
    missing_ok: bool = False

    def __init__(self, file: Union[Path, str], missing_ok: bool = False, **kwargs):
        self.file = file
        self.missing_ok = missing_ok
        super().__init__(**kwargs)

    def load(self) -> None:
        try:
            with open(self.file, self.mode) as fh:
                self.load_file_content(fh)
        except FileNotFoundError:
            if not self.missing_ok:
                raise

    @abstractmethod
    def load_file_content(self, handler: IO) -> None:
        pass


class PlainStructureParserMixin:
    delimiter: str = "_"
    prefix: str = "APP"
    data: Dict

    def save_key_value(self, key: str, value: Any) -> None:
        if self.prefix is None or key.startswith(self.prefix):
            levels, leaf = self.split_and_normalize(key)
            storage = self.leaf_level_storage(levels)
            storage[leaf] = value

    def leaf_level_storage(self, levels: List[str]) -> Dict:
        storage = self.data
        for level in levels:
            if level not in storage:
                storage[level] = {}
            storage = storage[level]
        return storage

    def split_and_normalize(self, key: str) -> Tuple[List[str], str]:
        parts = key.split(self.delimiter)
        normalized_parts = [p.lower() for p in filter(self.is_key_part, parts)]
        leaf = normalized_parts.pop()
        return normalized_parts, leaf

    def is_key_part(self, key: str) -> bool:
        if key:
            if self.prefix:
                return key != self.prefix
            return True
        return False


class Env(PlainStructureParserMixin, AbstractStorage):
    def __init__(self, delimiter="_", prefix="APP", **kwargs):
        self.delimiter = delimiter
        self.prefix = prefix
        super(Env, self).__init__(**kwargs)

    def load(self) -> None:
        for key, value in os.environ.items():
            self.save_key_value(key, value)


class DotEnv(PlainStructureParserMixin, BaseFileStorage):
    def __init__(
        self,
        file: Union[Path, str],
        prefix="APP",
        delimiter="_",
        missing_ok: bool = False,
        **kwargs
    ):
        self.delimiter = delimiter
        self.prefix = prefix
        super().__init__(file=file, missing_ok=missing_ok, **kwargs)

    def load_file_content(self, handler: IO) -> None:
        for line in filter(self.filter, handler.readlines()):
            key, value = self.split(line)
            self.save_key_value(key, value)

    @staticmethod
    def split(param_line: str) -> Tuple[str, str]:
        key, value = param_line.split("=", maxsplit=1)
        return key.strip(), value.strip()

    @staticmethod
    def filter(param_line: str) -> bool:
        return len(param_line) > 2 and (not param_line.startswith(("#", "//")))


class Json(BaseFileStorage):
    def load_file_content(self, handler: IO) -> None:
        self.data.update(json.load(handler))


class Ini(BaseFileStorage):
    def __init__(
        self,
        file: Union[Path, str],
        missing_ok=False,
        allow_no_value=False,
        delimiters=("=", ":"),
        comment_prefixes=("#", ";"),
        inline_comment_prefixes=None,
        strict=True,
        empty_lines_in_values=True,
        default_section=configparser.DEFAULTSECT,
        **kwargs
    ):
        self.parser = configparser.ConfigParser(
            allow_no_value=allow_no_value,
            delimiters=delimiters,
            comment_prefixes=comment_prefixes,
            inline_comment_prefixes=inline_comment_prefixes,
            strict=strict,
            empty_lines_in_values=empty_lines_in_values,
            default_section=default_section,
        )
        super().__init__(file=file, missing_ok=missing_ok, **kwargs)

    def load_file_content(self, handler: IO) -> None:
        self.parser.read_string(handler.read())
        self.data.update(self.parser)
